#!/usr/bin/env python3

import re
import os
import sys
import shutil
import subprocess
import json
from functools import partial

sys_re = re.compile("^/System")
usr_re = re.compile("^/usr/lib/")
exe_re = re.compile("@executable_path")

def is_user_lib(objfile, libname):
    return not sys_re.match(libname) and \
           not usr_re.match(libname) and \
           not exe_re.match(libname) and \
           not "libobjc." in libname and \
           not "libSystem." in libname and \
           not "libc." in libname and \
           not "libgcc." in libname and \
           not os.path.basename(libname) == 'Python' and \
           not os.path.basename(objfile) in libname and \
           not "libswift" in libname

def otool(objfile, rapths):
    command = "otool -L '%s' | grep -e '\t' | awk '{ print $1 }'" % objfile
    output  = subprocess.check_output(command, shell = True, universal_newlines=True)
    libs = set(filter(partial(is_user_lib, objfile), output.split()))

    libs_resolved = set()
    libs_relative = set()
    for lib in libs:
        lib_path = resolve_lib_path(objfile, lib, rapths)
        libs_resolved.add(lib_path)
        if lib_path != lib:
            libs_relative.add(lib)

    return libs_resolved, libs_relative

def get_rapths(objfile):
    rpaths = []
    command = "otool -l '%s' | grep -A2 LC_RPATH | grep path" % objfile
    pathRe = re.compile(r"^\s*path (.*) \(offset \d*\)$")

    try:
        result = subprocess.check_output(command, shell = True, universal_newlines=True)
    except:
        return rpaths

    for line in result.splitlines():
        line_clean = pathRe.search(line).group(1).strip()
        # resolve @loader_path
        if line_clean.startswith('@loader_path/'):
            line_clean = line_clean[len('@loader_path/'):]
            line_clean = os.path.normpath(os.path.join(os.path.dirname(objfile), line_clean))
        rpaths.append(line_clean)

    return rpaths

def get_rpaths_dev_tools(binary):
    command = "otool -l '%s' | grep -A2 LC_RPATH | grep path | grep \"Xcode\\|CommandLineTools\"" % binary
    result  = subprocess.check_output(command, shell = True, universal_newlines=True)
    pathRe = re.compile(r"^\s*path (.*) \(offset \d*\)$")
    output = []

    for line in result.splitlines():
        output.append(pathRe.search(line).group(1).strip())

    return output

def resolve_lib_path(objfile, lib, rapths):
    if os.path.exists(lib):
        return lib

    if lib.startswith('@rpath/'):
        lib = lib[len('@rpath/'):]
        for rpath in rapths:
            lib_path = os.path.join(rpath, lib)
            if os.path.exists(lib_path):
                return lib_path
    elif lib.startswith('@loader_path/'):
        lib = lib[len('@loader_path/'):]
        lib_path = os.path.normpath(os.path.join(objfile, lib))
        if os.path.exists(lib_path):
            return lib_path

    raise Exception('Could not resolve library: ' + lib)

def check_vulkan_max_version(version):
    try:
        result = subprocess.check_output("pkg-config vulkan --max-version=" + version, shell = True)
        return True
    except:
        return False

def get_homebrew_prefix():
    # set default to standard ARM path, intel path is already in the vulkan loader search array
    result = "/opt/homebrew"
    try:
        result = subprocess.check_output("brew --prefix", universal_newlines=True, shell=True, stderr=subprocess.DEVNULL).strip()
    except:
        pass

    return result

def install_name_tool_change(old, new, objfile):
    subprocess.call(["install_name_tool", "-change", old, new, objfile], stderr=subprocess.DEVNULL)

def install_name_tool_id(name, objfile):
    subprocess.call(["install_name_tool", "-id", name, objfile], stderr=subprocess.DEVNULL)

def install_name_tool_add_rpath(rpath, binary):
    subprocess.call(["install_name_tool", "-add_rpath", rpath, binary])

def install_name_tool_delete_rpath(rpath, binary):
    subprocess.call(["install_name_tool", "-delete_rpath", rpath, binary])

def libraries(objfile, result = dict(), result_relative = set(), rapths = []):
    rapths = get_rapths(objfile) + rapths
    libs_list, libs_relative = otool(objfile, rapths)
    result[objfile] = libs_list
    result_relative |= libs_relative

    for lib in libs_list:
        if lib not in result:
            libraries(lib, result, result_relative, rapths)

    return result, result_relative

def lib_path(binary):
    return os.path.join(os.path.dirname(binary), 'lib')

def resources_path(binary):
    return os.path.join(os.path.dirname(binary), '../Resources')

def lib_name(lib):
    return os.path.join("@executable_path", "lib", os.path.basename(lib))

def process_libraries(libs_dict, libs_dyn, binary):
    libs_set = set(libs_dict)
    # Remove binary from libs_set to prevent a duplicate of the binary being
    # added to the libs directory.
    libs_set.remove(binary)

    for src in libs_set:
        name = lib_name(src)
        dst = os.path.join(lib_path(binary), os.path.basename(src))

        shutil.copy(src, dst)
        os.chmod(dst, 0o755)
        install_name_tool_id(name, dst)

        if src in libs_dict[binary]:
            install_name_tool_change(src, name, binary)

        for p in libs_set:
            if p in libs_dict[src]:
                install_name_tool_change(p, lib_name(p), dst)

        for lib in libs_dyn:
            install_name_tool_change(lib, lib_name(lib), dst)

    for lib in libs_dyn:
        install_name_tool_change(lib, lib_name(lib), binary)

def process_swift_libraries(binary):
    command = ['xcrun', '--find', 'swift-stdlib-tool']
    swiftStdlibTool = subprocess.check_output(command, universal_newlines=True).strip()
    # from xcode11 on the dynamic swift libs reside in a separate directory from
    # the std one, might need versioned paths for future swift versions
    swiftLibPath = os.path.join(swiftStdlibTool, '../../lib/swift-5.0/macosx')
    swiftLibPath = os.path.abspath(swiftLibPath)

    command = [swiftStdlibTool, '--copy', '--platform', 'macosx', '--scan-executable', binary, '--destination', lib_path(binary)]

    if os.path.exists(swiftLibPath):
        command.extend(['--source-libraries', swiftLibPath])

    subprocess.check_output(command, universal_newlines=True)

    print(">> setting additional rpath for swift libraries")
    install_name_tool_add_rpath("@executable_path/lib", binary)

def process_vulkan_loader(binary, loaderName, loaderRelativeFolder, libraryNode):
    # https://github.com/KhronosGroup/Vulkan-Loader/blob/main/docs/LoaderDriverInterface.md#example-macos-driver-search-path
    # https://github.com/KhronosGroup/Vulkan-Loader/blob/main/docs/LoaderLayerInterface.md#macos-layer-discovery
    loaderSystemSearchFolders = [
        os.path.join(os.path.expanduser("~"), ".config", loaderRelativeFolder),
        os.path.join("/etc/xdg", loaderRelativeFolder),
        os.path.join("/usr/local/etc", loaderRelativeFolder),
        os.path.join("/etc", loaderRelativeFolder),
        os.path.join(os.path.expanduser("~"), ".local/share", loaderRelativeFolder),
        os.path.join("/usr/local/share", loaderRelativeFolder),
        os.path.join("/usr/share/vulkan", loaderRelativeFolder),
        os.path.join(get_homebrew_prefix(), 'share', loaderRelativeFolder),
    ]

    loaderSystemFolder = ""
    for loaderSystemSearchFolder in loaderSystemSearchFolders:
        if os.path.exists(loaderSystemSearchFolder):
            loaderSystemFolder = loaderSystemSearchFolder
            break

    if not loaderSystemFolder:
        print(">>> could not find loader folder " + loaderRelativeFolder)
        return

    loaderBundleFolder = os.path.join(resources_path(binary), loaderRelativeFolder)
    loaderSystemPath = os.path.join(loaderSystemFolder, loaderName)
    loaderBundlePath = os.path.join(loaderBundleFolder, loaderName)
    libraryRelativeFolder = "../../../Frameworks/"

    if not os.path.exists(loaderSystemPath):
        print(">>> could not find loader " + loaderName)
        return

    if not os.path.exists(loaderBundleFolder):
        os.makedirs(loaderBundleFolder)

    loaderSystemFile = open(loaderSystemPath, 'r')
    loaderJsonData = json.load(loaderSystemFile)
    librarySystemPath = os.path.join(loaderSystemFolder, loaderJsonData[libraryNode]["library_path"])

    if not os.path.exists(librarySystemPath):
        print(">>> could not find loader library " + librarySystemPath)
        return

    print(">>> modifiying and writing loader json " + loaderName)
    loaderBundleFile = open(loaderBundlePath, 'w')
    loaderLibraryName = os.path.basename(librarySystemPath)
    loaderJsonData[libraryNode]["library_path"] = os.path.join(libraryRelativeFolder, loaderLibraryName)
    json.dump(loaderJsonData, loaderBundleFile, indent=4)

    print(">>> copying loader library " + loaderLibraryName)
    frameworkBundleFolder = os.path.join(loaderBundleFolder, libraryRelativeFolder)
    if not os.path.exists(frameworkBundleFolder):
        os.makedirs(frameworkBundleFolder)
    shutil.copy(librarySystemPath, os.path.join(frameworkBundleFolder, loaderLibraryName))

def remove_dev_tools_rapths(binary):
    for path in get_rpaths_dev_tools(binary):
        install_name_tool_delete_rpath(path, binary)

def process(binary):
    binary = os.path.abspath(binary)
    if not os.path.exists(lib_path(binary)):
        os.makedirs(lib_path(binary))
    print(">> gathering all linked libraries")
    libs, libs_rel = libraries(binary)

    print(">> copying and processing all linked libraries")
    process_libraries(libs, libs_rel, binary)

    print(">> removing rpath definitions towards dev tools")
    remove_dev_tools_rapths(binary)

    print(">> copying and processing swift libraries")
    process_swift_libraries(binary)

    print(">> copying and processing vulkan loader")
    process_vulkan_loader(binary, "MoltenVK_icd.json", "vulkan/icd.d", "ICD")
    if check_vulkan_max_version("1.3.261.1"):
        process_vulkan_loader(binary, "VkLayer_khronos_synchronization2.json", "vulkan/explicit_layer.d", "layer")

if __name__ == "__main__":
    process(sys.argv[1])



#!/usr/bin/env python3

# This checks for the sdk path, the sdk version, and
# the sdk build version.

import re
import os
import string
import subprocess
import sys
from shutil import which
from subprocess import check_output

def find_macos_sdk():
    sdk = os.environ.get('MACOS_SDK', '')
    sdk_version = os.environ.get('MACOS_SDK_VERSION', '0.0')
    xcrun = which('xcrun')
    xcodebuild = which('xcodebuild')

    if not xcrun:
        return sdk,sdk_version

    if not sdk:
        sdk = check_output([xcrun, '--sdk', 'macosx', '--show-sdk-path'],
                            encoding="UTF-8")

    # find macOS SDK paths and version
    if sdk_version == '0.0':
        sdk_version = check_output([xcrun, '--sdk', 'macosx', '--show-sdk-version'],
                                    encoding="UTF-8")

        # use xcode tools when installed, still necessary for xcode versions <12.0
        try:
            sdk_version = check_output([xcodebuild, '-sdk', 'macosx', '-version', 'ProductVersion'],
                                        encoding="UTF-8", stderr=subprocess.DEVNULL)
        except:
            pass

    if not isinstance(sdk_version, str):
        sdk_version = '10.10.0'

    return sdk.strip(),sdk_version.strip()

if __name__ == "__main__":
    sdk_info = find_macos_sdk()
    sys.stdout.write(','.join(sdk_info))



#!/usr/bin/env python3
import os, sys, json, subprocess, re
from typing import Dict, Tuple, Callable, Optional

def call(cmd) -> str:
	sys.stdout.flush()
	ret = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, text=True)
	return ret.stdout

lint_rules: Dict[str, Tuple[Callable, str]] = {}

def lint_rule(description: str):
	def f(func):
		assert func.__name__ not in lint_rules.keys()
		lint_rules[func.__name__] = (func, description)
	return f

def get_commit_range() -> Optional[str]:
	if len(sys.argv) > 1:
		return sys.argv[1]
	# https://github.com/actions/runner/issues/342#issuecomment-590670059
	event_name = os.environ["GITHUB_EVENT_NAME"]
	with open(os.environ["GITHUB_EVENT_PATH"], "rb") as f:
		event = json.load(f)
	if event_name == "push":
		if event["created"] or event["forced"]:
			print("Skipping logic on branch creation or force-push")
			return None
		return event["before"] + "..." + event["after"]
	elif event_name == "pull_request":
		return event["pull_request"]["base"]["sha"] + ".." + event["pull_request"]["head"]["sha"]
	return None

def do_lint(commit_range: str) -> bool:
	commits = call(["git", "log", "--pretty=format:%h %s", commit_range]).splitlines()
	print(f"Linting {len(commits)} commit(s):")
	any_failed = False
	for commit in commits:
		sha, _, _ = commit.partition(' ')
		#print(commit)
		body = call(["git", "show", "-s", "--format=%B", sha]).splitlines()
		failed = []
		if len(body) == 0:
			failed.append("* Commit message must not be empty")
		else:
			for k, v in lint_rules.items():
				if not v[0](body):
					failed.append(f"* {v[1]} [{k}]")
		if failed:
			any_failed = True
			print("-" * 40)
			sys.stdout.flush()
			subprocess.run(["git", "-P", "show", "-s", sha])
			print("\nhas the following issues:")
			print("\n".join(failed))
			print("-" * 40)
	return any_failed

################################################################################

NO_PREFIX_WHITELIST = r"^Revert \"(.*)\"|^Reapply \"(.*)\"|^Release [0-9]|^Update VERSION$"

@lint_rule("Subject line must contain a prefix identifying the sub system")
def subsystem_prefix(body):
	if re.search(NO_PREFIX_WHITELIST, body[0]):
		return True
	m = re.search(r"^([^:]+):\s", body[0])
	if not m:
		return False
	# a comma-separated list is okay
	s = re.sub(r",\s+", "", m.group(1))
	# but no spaces otherwise
	return not " " in s

@lint_rule("First word after : must be lower case")
def description_lowercase(body):
	if re.search(NO_PREFIX_WHITELIST, body[0]):
		return True
	# Allow all caps for acronyms and such
	if re.search(r":\s[A-Z]{2,}\s", body[0]):
		return True
	return re.search(r":\s+[a-z0-9]", body[0])

@lint_rule("Subject line must not end with a full stop")
def no_dot(body):
	return not body[0].rstrip().endswith('.')

@lint_rule("There must be an empty line between subject and extended description")
def empty_line(body):
	return len(body) == 1 or body[1].strip() == ""

# been seeing this one all over github lately, must be the webshits
@lint_rule("Do not use 'conventional commits' style")
def no_cc(body):
	return not re.search(r"(?i)^(feat|fix|chore|refactor)[!:(]", body[0])

@lint_rule("History must be linear, no merge commits")
def no_merge(body):
	return not body[0].startswith("Merge ")

@lint_rule("Subject line should be shorter than 72 characters")
def line_too_long(body):
	revert = re.search(r"^Revert \"(.*)\"|^Reapply \"(.*)\"", body[0])
	return True if revert else len(body[0]) <= 72

@lint_rule("Prefix should not include C file extensions (use `vo_gpu: ...` not `vo_gpu.c: ...`)")
def no_file_exts(body):
	return not re.search(r"[a-z0-9]\.[ch]:\s", body[0])

################################################################################

if __name__ == "__main__":
	commit_range = get_commit_range()
	if commit_range is None:
		exit(0)
	print("Commit range:", commit_range)
	any_failed = do_lint(commit_range)
	exit(1 if any_failed else 0)
